/**********************************************************************
 *<
	FILE: SFCMODexp.cpp

	DESCRIPTION:  Starfleet Command Model file export module

	CREATED BY: Sean Dumas 

	HISTORY: created 16 December 1999

 *>	Copyright (c) 1999, All Rights Reserved.
 **********************************************************************/

#include <fstream.h>

#include "SFCMODImpExp.h"
#include "SFCMODExpResource.h"
#include "Max.h"
#include "istdplug.h"
#include "stdmat.h"
#include "decomp.h"
#include "helpsys.h"
#include "dummy.h"

const double kPI = 3.14159265;
const int vIdx[3] = { 0, 2, 1 };

class SceneEntry {
	public:
		TSTR name;
		INode *node,*tnode;
		Object *obj;
		int type;		// See above
		int id;
		SceneEntry *next;
		SceneEntry(INode *n, Object *o, int t);
		void SetID(int id) { this->id = id; }
};

void debugMsg( char *string, ... )
{
	FILE *debugFile;
	static bool firstTime = TRUE;
	static char outString[ 512 ];
	va_list argList;

#ifndef _DEBUG
	return;
#endif

	if ( firstTime )
	{
		debugFile = fopen( "debug.log", "wb" );
		firstTime = FALSE;
	}
	else
	{
		debugFile = fopen( "debug.log", "ab" );
	}

		// Get the variable argument list
	va_start( argList, string );
	vsprintf( outString, string, argList );
	va_end( argList );

	strcat( outString, "\r\n" );
	fprintf( debugFile, outString );

	fclose( debugFile );
}

//
// Beginning of old file
//

static BOOL showPrompts;
static BOOL exportSelected;

#define EXPORT_MTLS

#include <stdarg.h>

HINSTANCE hInstance;

TCHAR* GetString(int id)
{
    static TCHAR stBuf[ERROR_MSG_MAX_LEN];
	if (hInstance)
		return LoadString(hInstance, id, stBuf, ERROR_MSG_MAX_LEN) ? stBuf : NULL;
	return NULL;
}


static void MessageBox(int s1, int s2) {
	TSTR str1(GetString(s1));
	TSTR str2(GetString(s2));
	MessageBox(GetActiveWindow(), str1.data(), str2.data(), MB_OK);
	}

static int MessageBox(int s1, int s2, int option = MB_OK) {
	TSTR str1(GetString(s1));
	TSTR str2(GetString(s2));
	return MessageBox(GetActiveWindow(), str1, str2, option);
	}

static int Alert(int s1, int s2 = IDS_SD_SFCMODEXP, int option = MB_OK) {
	return MessageBox(s1, s2, option);
	}

class SceneEnumProc;
class MeshMtlList;
struct tPoint3
{
	float x, y, z;
};

static const int kMaxLOD = 5;
static float gHysteresis = 5;
static float gLODTransitions[kMaxLOD] = { 0, 100, 200, 300, 400 };

//////////////////////////////////////////////////////////////////////////////////////////
// ------------------------------ tSFCMODExport class --------------------------------
//

class tSFCMODExport : public SceneExport 
{
	friend BOOL CALLBACK ExportOptionsDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
	friend class tMesh;

private:
#pragma pack(1)
	struct tNamedPoint
	{
		tPoint3			fPos;
		unsigned int	fNameOffset;
	};

	struct tFaceDef
	{
		int vtxCount;
		int materialNdx;
		tPoint3 normal;
		struct tFlatPolyVertex
		{
			unsigned short vtxNdx;
			unsigned short flags;
			tPoint3 normal;
			float u;
			float v;
			float light;
		}vertices[3];
	};
#pragma pack()

	tNamedPoint* fNamedPoints;
	int fNumNamedPoints;

	Tab<int> fNameIndices;
	Tab<char> fStringPool;

	class tMesh
	{
	public:
		enum tSortKey
		{
			kSortByMaterialNdx,
			kSortByFlags
		};

		char*	fName;
		int		fNameSize;

		tPoint3*		fVertices;
		int				fNumVertices;
		tFaceDef*		fFaces;
		int				fNumFaces;
		
		tMesh* fNext;

		tMesh();
		~tMesh();
		void mCalcNormal(tFaceDef& face);
		static int mCompareFacesByMaterialNdx( const void* f1, const void* f2 );
		static int mCompareFacesByFlags( const void* f1, const void* f2 );
		void mMergeTriangles();
		void mSortFaces( tSortKey eSortKey );
		void mCalcVertexNormals( int polyGroupStart, int polyGroupStop );
	}*fFirstMesh;

	bool fAddGenericMaterial;
	unsigned int fSmoothedMaterials;

	bool fForceSmoothing;
	bool fCreateLODModel;

	void mGetLODInformation( SceneEnumProc& myScene, Interface* gi );
	void mGetNamedPoints( SceneEnumProc& myScene, Interface* gi );
	int mGetTextureNameIndex( int index );
	void mAddTextureNameIndex( int index );
	int mAddStringList( char * );
	void mListTextures( MeshMtlList& mtls );
	void mSortMeshes( const bool kComplexityName );
	void mGetMeshes( SceneEnumProc& myScene, Interface* gi, MeshMtlList& myMtls );
	float mGetBoundingRadius();
	int mSaveSFCMOD( const TCHAR* filename, ExpInterface* ei, Interface* gi );

public:
					tSFCMODExport();
					~tSFCMODExport();
	int				ExtCount();					// Number of extensions supported
    BOOL            MaxUVs;                     // TRUE if generating extra verts for mismatched UV coords
	const TCHAR *	Ext(int n);					// Extension #n (i.e. "3DS")
	const TCHAR *	LongDesc();					// Long ASCII description (i.e. "Autodesk 3D Studio File")
	const TCHAR *	ShortDesc();				// Short ASCII description (i.e. "3D Studio")
	const TCHAR *	AuthorName();				// ASCII Author name
	const TCHAR *	CopyrightMessage();			// ASCII Copyright message
	const TCHAR *	OtherMessage1();			// Other message #1
	const TCHAR *	OtherMessage2();			// Other message #2
	unsigned int	Version();					// Version number * 100 (i.e. v3.01 = 301)
	void			ShowAbout(HWND hWnd);		// Show DLL's "About..." box
	int				DoExport(const TCHAR *name,ExpInterface *ei,Interface *i, BOOL suppressPrompts, DWORD options);	// Export file
	BOOL			SupportsOptions(int ext, DWORD options);
};

tSFCMODExport::tMesh::tMesh()	:
	fVertices(NULL),
	fFaces(NULL),
	fNumVertices(0),
	fNumFaces(0),
	fNext(NULL)
{
}

tSFCMODExport::tMesh::~tMesh()
{
	if ( fVertices )
	{
		delete [] fVertices;
	}

	if ( fFaces )
	{
		delete [] fFaces;
	}
}

// Dialogue message proceedure
BOOL CALLBACK MsgDlgProc(HWND hDlg, UINT iMsg, UINT wParam, LONG lParam)
{
    switch (iMsg)
    {
	    case WM_INITDIALOG:
			CenterWindow(hDlg, GetWindow(hDlg, GW_OWNER));
			return TRUE;

	    case WM_COMMAND:
			switch(LOWORD(wParam))
            {
				case IDOK:	
					EndDialog(hDlg, TRUE); 
					break;
			}
			break;
	}
    return FALSE;
}

// Jaguar interface code

int controlsInit = FALSE;

BOOL WINAPI DllMain(HINSTANCE hinstDLL,ULONG fdwReason,LPVOID lpvReserved) 
{
	hInstance = hinstDLL;

	if ( !controlsInit ) {
		controlsInit = TRUE;
		
		// jaguar controls
		InitCustomControls(hInstance);

		// initialize Chicago controls
		InitCommonControls();
		}
	switch(fdwReason) {
		case DLL_PROCESS_ATTACH:
			//MessageBox(NULL,_T("3DSIMP.DLL: DllMain"),_T("3DSIMP"),MB_OK);
			break;
		case DLL_THREAD_ATTACH:
			break;
		case DLL_THREAD_DETACH:
			break;
		case DLL_PROCESS_DETACH:
			break;
		}
	return(TRUE);
}


//------------------------------------------------------

class tSFCMODClassDesc:public ClassDesc {
	public:
	int 			IsPublic() 
	{ 
		return 1; 
	}
	void *			Create(BOOL loading = FALSE) 
	{ 
		return new tSFCMODExport; 
	}
	const TCHAR *	ClassName() 
	{ 
		return GetString( IDS_SD_SFC ); 
	}
	SClass_ID		SuperClassID() 
	{ 
		return SCENE_EXPORT_CLASS_ID; 
	}
	Class_ID		ClassID() 
	{ 
		return Class_ID(0x64a03f4d, 0xd3a2549);
	}
	const TCHAR* 	Category() 
	{ 
		return GetString( IDS_SD_EXPORT );  
	}
};

static tSFCMODClassDesc gSFCMODDesc;

//------------------------------------------------------
// This is the interface to Jaguar:
//------------------------------------------------------

__declspec( dllexport ) const TCHAR *
LibDescription() 
{ 
	return GetString(IDS_SD_SFCMODEXPORTDLL ); 
}

__declspec( dllexport ) int
LibNumberClasses() 
{ 
	return 1; 
}

__declspec( dllexport ) ClassDesc *
LibClassDesc(int i) 
{
	switch(i) {		
		case 0: return &gSFCMODDesc; break;
		default: return 0; break;
		}
}

// Return version so can detect obsolete DLLs
__declspec( dllexport ) ULONG LibVersion() 
{ 
	return VERSION_3DSMAX; 
}

// Let the plug-in register itself for deferred loading
__declspec( dllexport ) ULONG CanAutoDefer()
{
	return 1;
}

//
// ._3DS export module functions follow:
//

void tSFCMODExport::tMesh::mCalcNormal(tFaceDef& face)
{
	Point3 p1( 
		fVertices[face.vertices[vIdx[0]].vtxNdx].x, 
		fVertices[face.vertices[vIdx[0]].vtxNdx].y, 
		fVertices[face.vertices[vIdx[0]].vtxNdx].z );
	Point3 p2( 
		fVertices[face.vertices[vIdx[1]].vtxNdx].x, 
		fVertices[face.vertices[vIdx[1]].vtxNdx].y, 
		fVertices[face.vertices[vIdx[1]].vtxNdx].z );
	Point3 p3( 
		fVertices[face.vertices[vIdx[2]].vtxNdx].x, 
		fVertices[face.vertices[vIdx[2]].vtxNdx].y, 
		fVertices[face.vertices[vIdx[2]].vtxNdx].z );

	Point3 n = CrossProd( p2 - p1, p3 - p1 );
	n = n.Normalize();

	face.normal.x = n.x;
	face.normal.y = n.y;
	face.normal.z = n.z;
}

void tSFCMODExport::tMesh::mCalcVertexNormals( int polyGroupStart, int polyGroupStop )
{
	int i;

	if ( fNumVertices < 1 )
	{
		return;
	}

	tPoint3* normals = new tPoint3[fNumVertices];
	int* counts = new int[fNumVertices];

	// Clear the arrays
	for ( i = 0; i < fNumVertices; i++ )
	{
		normals[i].x = 0;
		normals[i].y = 0;
		normals[i].z = 0;
		counts[i] = 0;
	}

	if ( polyGroupStop > fNumFaces )
	{
		polyGroupStop = fNumFaces;	// This shouldn't happen!
	}

	// Total neighboring normals
	for ( i = polyGroupStart; i < polyGroupStop; i++ )
	{
		for ( int j = 0; j < 3; j++ )
		{
			normals[fFaces[i].vertices[j].vtxNdx].x += fFaces[i].normal.x;
			normals[fFaces[i].vertices[j].vtxNdx].y += fFaces[i].normal.y;
			normals[fFaces[i].vertices[j].vtxNdx].z += fFaces[i].normal.z;
			counts[fFaces[i].vertices[j].vtxNdx]++;
		}
	}

	// average
	for ( i = 0; i < fNumVertices; i++ )
	{
		if ( counts[i] )
		{
			const float kScale = 1.0f / counts[i];
			normals[i].x *= kScale;
			normals[i].y *= kScale;
			normals[i].z *= kScale;
		}
	}

	// assign values
	for ( i = polyGroupStart; i < polyGroupStop; i++ )
	{
		for ( int j = 0; j < 3; j++ )
		{
			fFaces[i].vertices[j].normal = normals[fFaces[i].vertices[j].vtxNdx];
		}
	}

	delete [] normals;
	delete [] counts;
}

int tSFCMODExport::tMesh::mCompareFacesByMaterialNdx( const void* f1, const void* f2 )
{
	const tFaceDef* face1 = (const tFaceDef *) f1;
	const tFaceDef* face2 = (const tFaceDef *) f2;

	return face1->materialNdx - face2->materialNdx;
}

int tSFCMODExport::tMesh::mCompareFacesByFlags( const void* f1, const void* f2 )
{
	const tFaceDef* face1 = (const tFaceDef *) f1;
	const tFaceDef* face2 = (const tFaceDef *) f2;

	int flags1 = 0, flags2 = 0;

	if ( face1->vtxCount > 0 )
	{
		flags1 = (int) face1->vertices[0].flags;
	}

	if ( face2->vtxCount > 0 )
	{
		flags2 = (int) face2->vertices[0].flags;
	}

	return flags1 - flags2;
}

// Sort faces based on material index
void tSFCMODExport::tMesh::mSortFaces( tSFCMODExport::tMesh::tSortKey eSortKey )	
{
	switch ( eSortKey )
	{
	case kSortByMaterialNdx:
	default:
		qsort( fFaces, fNumFaces, sizeof(tFaceDef), mCompareFacesByMaterialNdx );
		break;
	case kSortByFlags:
		qsort( fFaces, fNumFaces, sizeof(tFaceDef), mCompareFacesByFlags );
		break;
	}
}

// Convert triangles to polygons and reduce the number faces
void tSFCMODExport::tMesh::mMergeTriangles()
{
}

tSFCMODExport::tSFCMODExport() :
	fFirstMesh(NULL),
	fNamedPoints(NULL),
	fNumNamedPoints(0),
	fAddGenericMaterial(false),
	fSmoothedMaterials(0),
	fForceSmoothing(true),
	fCreateLODModel(true)
{
}

tSFCMODExport::~tSFCMODExport() 
{
	if ( fNamedPoints )
	{
		delete [] fNamedPoints;
	}

	tMesh* mesh = fFirstMesh;
	while ( mesh )
	{
		tMesh* oldMesh = mesh;
		mesh = mesh->fNext;
		delete oldMesh;
	}
}

int tSFCMODExport::ExtCount() 
{
	return 1;
}

const TCHAR* tSFCMODExport::Ext(int n) 
{		
	// Extensions supported for import/export modules
	switch(n) {
		case 0:
			return _T("MOD");
		}
	return _T("");
}

const TCHAR* tSFCMODExport::LongDesc() 
{
	// Long ASCII description (i.e. "Targa 2.0 Image File")
	return GetString(IDS_SD_SFCMODFILE);
}
	
const TCHAR* tSFCMODExport::ShortDesc() 
{
	// Short ASCII description (i.e. "Targa")
	return GetString(IDS_SD_SFCFILE);
}

const TCHAR* tSFCMODExport::AuthorName() 
{
	// ASCII Author name
	return GetString(IDS_SD_AUTHOR);	// Hey, that's me!!!!!
}

const TCHAR* tSFCMODExport::CopyrightMessage() 
{	
	// ASCII Copyright message
	return GetString(IDS_SD_COPYRIGHT);
}

const TCHAR* tSFCMODExport::OtherMessage1() 
{		// Other message #1
	return _T("");
}

const TCHAR* tSFCMODExport::OtherMessage2() 
{		// Other message #2
	return _T("");
}

unsigned int tSFCMODExport::Version() 
{				// Version number * 100 (i.e. v3.01 = 301)
	return 200;
}

void tSFCMODExport::ShowAbout(HWND hWnd) 
{			// Optional
}

#define OBTYPE_MESH 0
#define OBTYPE_CAMERA 1
#define OBTYPE_OMNILIGHT 2
#define OBTYPE_SPOTLIGHT 3
#define OBTYPE_DUMMY 5
#define OBTYPE_CTARGET 6
#define OBTYPE_LTARGET 7

SceneEntry::SceneEntry(INode *n, Object *o, int t) 
{ 
	node = n; obj = o; type = t; next = NULL; 
	tnode = n->GetTarget();
}


//- Material Export -------------------------------------------------------------

struct MEntry { Mtl *m; };

class MeshMtlList: public Tab<MEntry> 
{
public:
	void AddMtl(Mtl *m);
	void ReallyAddMtl(Mtl *m);
	int FindMtl(Mtl *m);
	int FindSName(char *nam);
	~MeshMtlList() 
	{
		for (int i=0; i<Count(); i++) 
		{
		}
	}
};

int MeshMtlList::FindMtl(Mtl *m) 
{
	for (int i=0; i<Count(); i++) 
	{
		if  ( (*this)[i].m == m ) 
		{
			return i;
		}
	}
	return -1;
}

static int IsStdMulti(Mtl *m) 
{
	return (m->ClassID()==Class_ID(MULTI_CLASS_ID,0))?1:0; 
}

int  MeshMtlList::FindSName(char *name) 
{
	for (int i=0; i<Count(); i++) 
	{
	}
	return -1;
}

void MeshMtlList::AddMtl(Mtl *m) 
{
	debugMsg( "MeshMtlList::AddMtl( Mtl* (%h) )", m );
	if (m==NULL) return;
	Interval v;
	m->Update(0,v);
	if (IsStdMulti(m)) 
	{
		for (int i=0; i<m->NumSubMtls(); i++) 
		{
			Mtl *sub = m->GetSubMtl(i);
			if (sub&&FindMtl(sub)<0) 
			{
				ReallyAddMtl(sub);
			}
		}
	}
	else 
	{
		if (FindMtl(m)<0) 
		{
			ReallyAddMtl(m);
		}
	}
}

static void IncrName(char *name, char *inm, int n) 
{
	char buf[20];
	strncpy(buf,name,13);
	buf[13] = 0;
	sprintf(inm,"%s%d\0",buf,n);
}

void MeshMtlList::ReallyAddMtl(Mtl *m) 
{
	if ( m )
	{
		debugMsg( "MeshMtlList::ReallyAddMtl() --> %s", m->GetName() );
	}
	MEntry me;
	me.m = m;
	char buf[20];
	strncpy(buf,me.m->GetName(),16);
	if (strlen(buf)==0) strcpy(buf, "Matl");
	buf[16] = 0;
	int n=0;
	while (FindSName(m->GetName())>=0) 
		IncrName(buf,m->GetName(),++n); 
	buf[16] = 0;
	Append(1,&me,20);
}								

//--------------------------------------------------------------

class SceneEnumProc : public ITreeEnumProc 
{
	public:
		Interface	*i;
		SceneEntry *head;
		SceneEntry *tail;
		IScene		*theScene;
		int			count;
		MeshMtlList *mtlList;
		TimeValue	time;
					SceneEnumProc(IScene *scene, TimeValue t, Interface *i, MeshMtlList *ml);
					~SceneEnumProc();
		int			Count() { return count; }
		void		Append(INode *node, Object *obj, int type);
		int			callback( INode *node );
		Box3		Bound();
		SceneEntry *Find(INode *node);
//		SceneEntry *operator[](int index);
};

SceneEnumProc::SceneEnumProc(IScene *scene, TimeValue t, Interface *i, MeshMtlList *ml) 
{
	time = t;
	theScene = scene;
	count = 0;
	head = tail = NULL;
	this->i = i;
	mtlList = ml;
	theScene->EnumTree(this);
}

SceneEnumProc::~SceneEnumProc() 
{
	while(head) 
	{
		SceneEntry *next = head->next;
		delete head;
		head = next;
	}
	head = tail = NULL;
	count = 0;	
}

int SceneEnumProc::callback(INode *node) 
{
	if(exportSelected && node->Selected() == FALSE)
	{
		return TREE_CONTINUE;
	}
	Object *obj = node->EvalWorldState(time).obj;
	if (obj->CanConvertToType(triObjectClassID)) 
	{
		Append(node, obj, OBTYPE_MESH);
		mtlList->AddMtl(node->GetMtl());
		return TREE_CONTINUE;
	}
	if (node->IsTarget()) 
	{
		INode* ln = node->GetLookatNode();
		if (ln) 
		{
			Object *lobj = ln->EvalWorldState(time).obj;
			switch(lobj->SuperClassID()) 
			{
				case LIGHT_CLASS_ID:  Append(node, obj, OBTYPE_LTARGET); break;
				case CAMERA_CLASS_ID: Append(node, obj, OBTYPE_CTARGET); break;
			}
		}
		return TREE_CONTINUE;
	}
	switch (obj->SuperClassID()) 
	{ 
		case HELPER_CLASS_ID:
			if ( obj->ClassID()==Class_ID(DUMMY_CLASS_ID,0)) 
				Append(node, obj, OBTYPE_DUMMY);
			break;
		case LIGHT_CLASS_ID: 
		{
            TCHAR buf[ERROR_MSG_MAX_LEN];
			if (obj->ClassID()==Class_ID(OMNI_LIGHT_CLASS_ID,0))
				Append(node, obj, OBTYPE_OMNILIGHT);
			if (obj->ClassID()==Class_ID(SPOT_LIGHT_CLASS_ID,0)) 
				Append(node, obj, OBTYPE_SPOTLIGHT);
            if (obj->ClassID()==Class_ID(DIR_LIGHT_CLASS_ID,0)) 
			{
				debugMsg( "DIR_LIGHT_CLASS_ID" );
//                sprintf(buf, GetString(/*IDS_DIR_LIGHT*/0), node->GetName());
//		        AddToMsgList(msgList, buf);
				//Append(node, obj, OBTYPE_SPOTLIGHT);
            }
            if (obj->ClassID()==Class_ID(FSPOT_LIGHT_CLASS_ID,0)) 
			{
				debugMsg( "FSPOT_LIGHT_CLASS_ID" );
//                sprintf(buf, GetString(/*IDS_FSPOT_LIGHT*/0), node->GetName());
//		        AddToMsgList(msgList, buf);
				//Append(node, obj, OBTYPE_SPOTLIGHT);
            }
            break;
        }
		case CAMERA_CLASS_ID:
			if (obj->ClassID()==Class_ID(LOOKAT_CAM_CLASS_ID,0))
				Append(node, obj, OBTYPE_CAMERA);
			break;
	}
	return TREE_CONTINUE;	// Keep on enumeratin'!
}


void SceneEnumProc::Append(INode *node, Object *obj, int type) 
{
	SceneEntry *entry = new SceneEntry(node, obj, type);

	if(tail)
		tail->next = entry;
	tail = entry;
	if(!head)
		head = entry;
	count++;	
}

Box3 SceneEnumProc::Bound() 
{
	Box3 bound;
	bound.Init();
	SceneEntry *e = head;
	ViewExp *vpt = i->GetViewport(NULL);
	while(e) 
	{
		Box3 bb;
		e->obj->GetWorldBoundBox(time, e->node, vpt, bb);
		bound += bb;
		e = e->next;
	}
	return bound;
}

SceneEntry *SceneEnumProc::Find(INode *node) 
{
	SceneEntry *e = head;
	while(e) 
	{
		if(e->node == node)
			return e;
		e = e->next;
	}
	return NULL;
}


SceneEnumProc *theSceneEnum = NULL;
MeshMtlList *theMtls = NULL;

class tChunk
{
private:
	FILE* fFile;
	int fPos;


public:
	void mStart( FILE* _file, FOURCC _name );
	void mFixUp();
};

void tChunk::mStart( FILE* _file, FOURCC _name )
{
	fFile = _file;
	fwrite( &_name, sizeof(FOURCC), 1, fFile );	
	int offset = 0;
	fwrite( &offset, sizeof(int), 1, fFile );	// Reserve space for the offset value
	fpos_t pos_t;
	fgetpos( fFile, &pos_t );
	fPos = (int) pos_t;
}

void tChunk::mFixUp()
{
	fpos_t pos;
	fgetpos( fFile, &pos );
	fseek( fFile, fPos - sizeof(int), SEEK_SET );
	int offset = pos - fPos;
	fwrite( &offset, sizeof(int), 1, fFile );
	fseek( fFile, 0, SEEK_END );

	// If we need to pad to keep dword alignment, do so now
	//
	for ( int pad = ( 4 - ( pos & 3 ) ) & 3; pad; pad--)
	{
		fputc( 0, fFile );
	}
}

void tSFCMODExport::mAddTextureNameIndex( int index )
{
	fNameIndices.Append( 1, &index );
}

int tSFCMODExport::mGetTextureNameIndex( int index )
{
	return fNameIndices[index];
}

int tSFCMODExport::mAddStringList( char* _string )
{
	CStr string( _string );
	int offset = 0;
	string.toLower();	// $$$

	// first search for the string in the list
	for ( int i = 0; i < fStringPool.Count(); )
	{
		CStr pool( fStringPool.Addr(i) );
		if ( pool == string )
		{
			return i;	// Got one!!!!!!!!
		}
		else
		{
			i += strlen( fStringPool.Addr(i) ) + 1;
			offset = i;
		}
	}

	// Append the new string

	int size = fStringPool.Count() + string.length() + 1;
	fStringPool.Resize( size );
	fStringPool.SetCount( size );
	strcpy( fStringPool.Addr(offset), string.data() );
	(*fStringPool.Addr( offset + string.length() )) = NULL;

	return offset;
}

void tSFCMODExport::mListTextures( MeshMtlList& mtls )
{
	int i;
	for ( i = 0; i < mtls.Count(); i++ )
	{
		// Debug the Max material structures
		if( mtls[i].m && mtls[i].m->GetName().data() )
		{
			for ( int j = 0; j < mtls[i].m->NumSubTexmaps(); j++ )
			{
				Texmap* map = mtls[i].m->GetSubTexmap(j);
				if ( map && ( j == ID_DI || j == ID_SI ) )	// Get diffuse and self-illumination
				{
					// just do bitmap textures for now
					BitmapTex* bitmap = NULL;
					if ( map->ClassID() == Class_ID( BMTEX_CLASS_ID, 0 ) )
					{
						bitmap = (BitmapTex *) map;
						if ( map->GetName().data() && bitmap->GetMapName() )
						{
							debugMsg( "%s", bitmap->GetMapName() );
							static char filename[_MAX_FNAME];
							_splitpath( bitmap->GetMapName(), NULL, NULL, filename, NULL );
							debugMsg( "Adding texture: %s from %s", filename, bitmap->GetMapName() );
							mAddTextureNameIndex( mAddStringList(filename) );
							debugMsg( "done->mAddTextureNameIndex" );
						}
					}
				}
			}
		}
	}
}

// Determine is the node has negative scaling.
// This is used for mirrored objects for example. They have a negative scale factor
// so when calculating the normal we should take the vertices counter clockwise.
// If we don't compensate for this the objects will be 'inverted'.
static bool NegParity(Matrix3 &m)
{
	return (DotProd(CrossProd(m.GetRow(0),m.GetRow(1)),m.GetRow(2))<0.0) ? true : false;
}

void tSFCMODExport::mGetNamedPoints( SceneEnumProc& myScene, Interface* gi )
{
	if ( fNamedPoints )
	{
		delete [] fNamedPoints;
	}

    SceneEntry* pSceneEntry = myScene.head;

	// First count the number of dummys in the scene

	fNumNamedPoints = 0;
	while ( pSceneEntry )    
    { 
		Object *pObject = pSceneEntry->obj;
		if ( pObject->ClassID() == Class_ID(DUMMY_CLASS_ID,0) )
		{
			fNumNamedPoints++;
		}
		pSceneEntry = pSceneEntry->next;
	}

	// Now get the origins

	if ( !fNumNamedPoints )
	{
		return;
	}

	fNamedPoints = new tNamedPoint[fNumNamedPoints];
	int index = 0;
	pSceneEntry = myScene.head;
	while ( pSceneEntry )    
    { 
		Object *pObject = pSceneEntry->obj;
        INode *pNode = pSceneEntry->node;
        TCHAR *pNodeName = pNode->GetName();
		Matrix3 transMatrix = pNode->GetObjTMAfterWSM(gi->GetTime());
		transMatrix.RotateX(-kPI*0.5f);
		transMatrix.RotateY(kPI*0.5f);
		if ( pObject->ClassID() == Class_ID(DUMMY_CLASS_ID,0) )
		{
//			DummyObject* pDummy = (DummyObject *) pObject;
			Point3 v = Point3(0,0,0) * transMatrix;
			fNamedPoints[index].fNameOffset = mAddStringList(pNodeName);
			fNamedPoints[index].fPos.x = v.x;
			fNamedPoints[index].fPos.y = v.y;
			fNamedPoints[index++].fPos.z = v.z;
		}
		pSceneEntry = pSceneEntry->next;
	}
}

void tSFCMODExport::mGetLODInformation( SceneEnumProc& myScene, Interface* gi )
{
    SceneEntry* pSceneEntry = NULL;

	if ( fCreateLODModel )
	{
		pSceneEntry = myScene.head;
		for ( pSceneEntry = myScene.head; pSceneEntry; pSceneEntry = pSceneEntry->next )
		{
			Object* pObject = pSceneEntry->obj;
			INode* pNode = pSceneEntry->node;
			if ( pObject->ClassID() == Class_ID( SFCLOD_CLASS_ID1, SFCLOD_CLASS_ID2 ) )
			{
				TSTR string;
				if ( pNode->GetUserPropString( "LOD", string ) )	// $$$ could overrun the array...
				{
					char* token;
					token = strtok( string.data(), " " );
					gHysteresis = atoi( token );
					token = strtok( NULL, " " );
					for ( int i = 0; i < kMaxLOD && token; i++ )
					{
						gLODTransitions[i] = atof(token);
						token = strtok( NULL, " " );
					}
				}
				break;
			}
		}
	}
}

void tSFCMODExport::mGetMeshes( SceneEnumProc& myScene, Interface* gi, MeshMtlList& myMtls )
{
    SceneEntry* pSceneEntry = NULL;

	if ( !fCreateLODModel )
	{
		// Create a single mesh
		// Count the number of polygons and  vertices
		fFirstMesh = new tMesh;
		pSceneEntry = myScene.head;
		for ( pSceneEntry = myScene.head; pSceneEntry; pSceneEntry = pSceneEntry->next )
		{
			Object* pObject = pSceneEntry->obj;
			if ( pObject->CanConvertToType(triObjectClassID) )
			{
				TriObject* pTriangle = (TriObject *) pObject->ConvertToType( gi->GetTime(), triObjectClassID );
				if ( pTriangle )
				{
					Mesh& mesh = pTriangle->GetMesh();
					fFirstMesh->fNumVertices += mesh.getNumVerts();
					fFirstMesh->fNumFaces += mesh.getNumFaces();
				}
			}
		}
		if ( fFirstMesh->fNumVertices )
		{
			fFirstMesh->fVertices = new tPoint3[fFirstMesh->fNumVertices];
		}
		if ( fFirstMesh->fNumFaces )
		{
			fFirstMesh->fFaces = new tFaceDef[fFirstMesh->fNumFaces];
		}
	}

	int polyBase = 0;
	int vertexBase = 0;

    int group = 0;
	pSceneEntry = myScene.head;
	while ( pSceneEntry )    
    { 
        INode *pNode = pSceneEntry->node;
        TCHAR *pNodeName = pNode->GetName();

		// Match all the materials with those in our list
		Mtl* pMasterMtl = pNode->GetMtl();
		int* materialIndices;
		int numObjMtls = 0;
		if ( pMasterMtl && pMasterMtl->NumSubMtls() > 0 )
		{
			numObjMtls = pMasterMtl->NumSubMtls();
			materialIndices = new int[numObjMtls];
			for ( int i = 0; i < numObjMtls; i++ )
			{
				if ( pMasterMtl->GetSubMtl(i) )
				{
					debugMsg( "Getting sub material %d, %s", i, pMasterMtl->GetSubMtl(i)->GetName() );
				}
				materialIndices[i] = myMtls.FindMtl( pMasterMtl->GetSubMtl(i) );
				if ( materialIndices[i] < 0 )
				{
					fAddGenericMaterial = true;
					materialIndices[i] = -1;
				}
			}
		}
		else if ( pMasterMtl )
		{
			debugMsg( "Getting material %s", pMasterMtl->GetName() );
			numObjMtls = 1;
			materialIndices = new int[1];
			materialIndices[0] = myMtls.FindMtl( pMasterMtl );
			if ( materialIndices[0] < 0 )
			{
				fAddGenericMaterial = true;
				materialIndices[0] = -1;
			}
		}
		else
		{
			fAddGenericMaterial = true;
			numObjMtls = 1;
			materialIndices = new int[1];			
			materialIndices[0] = -1;
		}

		Object *pObject = pSceneEntry->obj;
		Matrix3 transMatrix = pNode->GetObjTMAfterWSM(gi->GetTime());
		transMatrix.RotateX(-kPI*0.5f);
		transMatrix.RotateY(kPI*0.5f);

		////////////////////////
		// Fill the vertices
		//
		tMesh* pMesh = NULL;
		if ( pObject->CanConvertToType(triObjectClassID) )
		{
			// Create a new mesh
			tMesh* pMesh;
			if ( fCreateLODModel )
			{
				pMesh = new tMesh;
				pMesh->fNext = fFirstMesh;
				fFirstMesh = pMesh;
			}
			else
			{
				pMesh = fFirstMesh;
			}

			if ( fCreateLODModel )
			{
				pMesh->fNameSize = strlen( pNode->GetName() ) + 1;
				if ( pMesh->fNameSize )
				{
					pMesh->fName = new char[pMesh->fNameSize];
					strcpy( pMesh->fName, pNode->GetName() );
				}
			}

			TriObject* pTriangle = (TriObject *) pObject->ConvertToType( gi->GetTime(), triObjectClassID );
			if ( pTriangle )
			{
				Mesh& mesh = pTriangle->GetMesh();	
				if ( fCreateLODModel )
				{
					pMesh->fNumVertices = mesh.getNumVerts();
					pMesh->fVertices = new tPoint3[mesh.getNumVerts()];
				}
				for ( int i = 0; i < mesh.getNumVerts(); i++ )
				{
					Point3 v = mesh.verts[i] * transMatrix;
					pMesh->fVertices[vertexBase + i].x = v.x;
					pMesh->fVertices[vertexBase + i].y = v.y;
					pMesh->fVertices[vertexBase + i].z = v.z;
				}

				//////////////////////
				// Fill the faces
				//

				// Save the faces
				if ( fCreateLODModel )
				{
					pMesh->fNumFaces = mesh.getNumFaces();
					pMesh->fFaces = new tFaceDef[mesh.getNumFaces()];
				}
				for ( i = 0; i < mesh.getNumFaces(); i++ )
				{
					int mtlNdx = mesh.getFaceMtlIndex(i);
					if ( mtlNdx >= numObjMtls && numObjMtls )
					{
						mtlNdx %= numObjMtls;
					}
					else if ( mtlNdx >= numObjMtls )
					{
						mtlNdx = 0;
					}

					memset( &pMesh->fFaces[polyBase + i], 0, sizeof(tFaceDef) );
					pMesh->fFaces[polyBase + i].vtxCount = 3;
					pMesh->fFaces[polyBase + i].materialNdx = materialIndices[mtlNdx];
					if ( mesh.faces[i].smGroup && materialIndices[mtlNdx] >= 0 )	// Primitive
					{
						fSmoothedMaterials |= 1 << pMesh->fFaces[polyBase + i].materialNdx;
					}
					for ( int j = 0; j < 3; j++ )
					{
						if ( !fCreateLODModel )
						{
							pMesh->fFaces[polyBase + i].vertices[j].flags = group;
						}
						if ( mesh.tVerts )
						{
							pMesh->fFaces[polyBase + i].vertices[j].u = mesh.tVerts[mesh.tvFace[i].t[vIdx[j]]].x;
							pMesh->fFaces[polyBase + i].vertices[j].v = mesh.tVerts[mesh.tvFace[i].t[vIdx[j]]].y;
						}
						pMesh->fFaces[polyBase + i].vertices[j].vtxNdx = (unsigned short) mesh.faces[i].v[vIdx[j]] + vertexBase;
					}
					pMesh->mCalcNormal(pMesh->fFaces[polyBase + i]);
				}

				if ( !fCreateLODModel )
				{
					polyBase += mesh.getNumFaces();
					vertexBase += mesh.getNumVerts();
				}
			}


			if ( fCreateLODModel )
			{
				debugMsg( "Reduce Triangles, convert to polygons" );
				pMesh->mMergeTriangles();
				debugMsg( "Sorting faces..." );
				pMesh->mSortFaces(tMesh::kSortByMaterialNdx);
				debugMsg( "Calculating Normals..." );
				if ( fForceSmoothing )
				{
					pMesh->mCalcVertexNormals( 0, pMesh->fNumFaces );
				}
				else
				{				
					for ( int i = 0; i < pMesh->fNumFaces; )
					{
						for ( int j = i; j < pMesh->fNumFaces && pMesh->fFaces[j].materialNdx == pMesh->fFaces[i].materialNdx; j++ )
						{
						}
						pMesh->mCalcVertexNormals( i, j );
						i = j;
					}
				}
			}
			else
			{
				group++;
			}
		}
		delete [] materialIndices;
		pSceneEntry = pSceneEntry->next;
	}

	if ( !fCreateLODModel && fFirstMesh )
	{
			debugMsg( "Reduce Triangles, convert to polygons" );
			fFirstMesh->mMergeTriangles();
			debugMsg( "Sorting faces..." );
			fFirstMesh->mSortFaces(tMesh::kSortByMaterialNdx);
			debugMsg( "Calculating Normals..." );
			if ( fForceSmoothing )
			{
				fFirstMesh->mCalcVertexNormals( 0, fFirstMesh->fNumFaces );
			}
			else
			{
				for ( int i = 0; i < fFirstMesh->fNumFaces; )
				{
					for ( int j = i; j < fFirstMesh->fNumFaces && fFirstMesh->fFaces[j].materialNdx == fFirstMesh->fFaces[i].materialNdx; j++ )
					{
					}
					fFirstMesh->mCalcVertexNormals( i, j );
					i = j;
				}
			}
			debugMsg( "Sorting faces again..." );
			fFirstMesh->mSortFaces(tMesh::kSortByFlags);
	}

	// Correct the material indices
	if ( fAddGenericMaterial )
	{
		tMesh* pMesh = fFirstMesh;
		while ( pMesh )
		{
			for ( int i = 0; i < pMesh->fNumFaces; i++ )
			{
				pMesh->fFaces[i].materialNdx++;
			}
			pMesh = pMesh->fNext;
		}
	}
}

void tSFCMODExport::mSortMeshes( const bool kComplexityName )
{
	// Sort the meshes based on complexity
	tMesh* pSorted = NULL;
	while ( fFirstMesh )
	{
		tMesh* pPrevious = NULL;
		tMesh* pSortedMesh = pSorted;
		tMesh* pMesh = fFirstMesh;
		fFirstMesh = fFirstMesh->fNext;
		while ( pSortedMesh )
		{
			if ( kComplexityName )
			{
				if ( pSortedMesh->fNumFaces <= pMesh->fNumFaces )
				{
					break;
				}
			}
			else
			{
				if ( stricmp( pSortedMesh->fName, pMesh->fName ) >= 0 )
				{
					break;
				}
			}
			pPrevious = pSortedMesh;
			pSortedMesh = pSortedMesh->fNext;
		}
		if ( pPrevious )
		{
			pPrevious->fNext = pMesh;
		}
		pMesh->fNext = pSortedMesh;
		if ( pSorted == pSortedMesh )
		{
			pSorted = pMesh;
		}
	}
	fFirstMesh = pSorted;
}

static bool gComplexityName = true;
static BOOL CALLBACK LODDialogProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) 
{
	static TSTR name;
	char string[40];

	switch (msg) {
	case WM_INITDIALOG:
		CheckRadioButton( hWnd, IDC_COMPLEXITY, IDC_NAME, gComplexityName ? IDC_COMPLEXITY : IDC_NAME );
		sprintf( string, "%6.2f", gHysteresis );
		SetWindowText( GetDlgItem(hWnd,IDC_HYSTERESIS), string );
		sprintf( string, "%6.2f", gLODTransitions[0] );
		SetWindowText( GetDlgItem(hWnd,IDC_LOD1), string );
		sprintf( string, "%6.2f", gLODTransitions[1] );
		SetWindowText( GetDlgItem(hWnd,IDC_LOD2), string );
		sprintf( string, "%6.2f", gLODTransitions[2] );
		SetWindowText( GetDlgItem(hWnd,IDC_LOD3), string );
		sprintf( string, "%6.2f", gLODTransitions[3] );
		SetWindowText( GetDlgItem(hWnd,IDC_LOD4), string );
		sprintf( string, "%6.2f", gLODTransitions[4] );
		SetWindowText( GetDlgItem(hWnd,IDC_LOD5), string );
		CenterWindow(hWnd, GetParent(hWnd));
		SetFocus(GetDlgItem(hWnd,IDC_HYSTERESIS));
		return FALSE;

	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case IDOK:
	        gComplexityName = IsDlgButtonChecked( hWnd, IDC_COMPLEXITY );

			name.Resize (GetWindowTextLength(GetDlgItem(hWnd,IDC_HYSTERESIS))+1);
			GetWindowText (GetDlgItem(hWnd,IDC_HYSTERESIS), name.data(), name.length()+1);
			gHysteresis = atoi( name.data() );

			name.Resize (GetWindowTextLength(GetDlgItem(hWnd,IDC_LOD1))+1);
			GetWindowText (GetDlgItem(hWnd,IDC_LOD1), name.data(), name.length()+1);
			gLODTransitions[0] = atof( name.data() );

			name.Resize (GetWindowTextLength(GetDlgItem(hWnd,IDC_LOD2))+1);
			GetWindowText (GetDlgItem(hWnd,IDC_LOD2), name.data(), name.length()+1);
			gLODTransitions[1] = atof( name.data() );

			name.Resize (GetWindowTextLength(GetDlgItem(hWnd,IDC_LOD3))+1);
			GetWindowText (GetDlgItem(hWnd,IDC_LOD3), name.data(), name.length()+1);
			gLODTransitions[2] = atof( name.data() );

			name.Resize (GetWindowTextLength(GetDlgItem(hWnd,IDC_LOD4))+1);
			GetWindowText (GetDlgItem(hWnd,IDC_LOD4), name.data(), name.length()+1);
			gLODTransitions[3] = atof( name.data() );

			name.Resize (GetWindowTextLength(GetDlgItem(hWnd,IDC_LOD5))+1);
			GetWindowText (GetDlgItem(hWnd,IDC_LOD5), name.data(), name.length()+1);
			gLODTransitions[4] = atof( name.data() );

			EndDialog(hWnd,1);
			break;

		case IDCANCEL:
			EndDialog(hWnd,0);
			break;
		}
		break;

	default:
		return 0;
	}
	return 1;
}

float tSFCMODExport::mGetBoundingRadius()
{
	int i;

	// Find bounding coordinants for all the meshes
	tMesh* pMesh = fFirstMesh;
	float longestVector = 0;
	while ( pMesh )
	{
		for ( i = 0; i < pMesh->fNumVertices; i++ )
		{
			float length = 
				pMesh->fVertices[i].x * pMesh->fVertices[i].x +
				pMesh->fVertices[i].y * pMesh->fVertices[i].y +
				pMesh->fVertices[i].z * pMesh->fVertices[i].z;
			if ( length > longestVector )
			{
				longestVector = length;
			}
		}
		pMesh = pMesh->fNext;
	}

	return sqrt(longestVector);
}

int tSFCMODExport::mSaveSFCMOD( const TCHAR* filename, ExpInterface* ei, Interface* gi )
{
	int i;

	// Open a file stream
	FILE* myFile = fopen( filename, "wb" );

    MeshMtlList myMtls;

	debugMsg( "myScene" );
	// Ask the scene to enumerate all its nodes so we can determine if there are any we can use
	SceneEnumProc myScene( ei->theScene, gi->GetTime(), gi, &myMtls );

	// Get Named points
	debugMsg( "mGetNamedPoints" );
	mGetNamedPoints( myScene, gi );

	// Prompt for force smoothing across the entire model
	fForceSmoothing = MessageBox( IDS_SD_FORCESMOOTHINGPROMPT, IDS_SFCMODEXP, MB_YESNO ) == IDYES;

	// Prompt for Regular Model or Break-up model
	fCreateLODModel = MessageBox( IDS_SD_LODMODELPROMPT, IDS_SFCMODEXP, MB_YESNO ) == IDYES;

	// Get meshes
	debugMsg( "mGetMeshes" );
	mGetMeshes( myScene, gi, myMtls );

//	mAddStringList( "$$$FILLER$$$" );	// Add a place holder string to prevent error in the model viewer
	debugMsg( "mListTextures" );
	mListTextures( myMtls );

	debugMsg( "MDLS chunk" );
	///////////////////////
	// Main MDLS chunk
	//

	// Count the number of models
	tMesh* pMesh = fFirstMesh;
	for ( int modelCount = 0; pMesh; modelCount++ )
	{
		pMesh = pMesh->fNext;
	}

	// If modelCount is greater than 1 we need LOD information from the user
	if ( modelCount > 1 )
	{
		DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_LOD), gi->GetMAXHWnd(), LODDialogProc, NULL);
	}

	mSortMeshes( gComplexityName );

	mGetLODInformation( myScene, gi );

	tChunk mainChunk;
	mainChunk.mStart( myFile, MAKEFOURCC( 'M', 'D', 'L', 'S' ) );
	const int kVerHi = 2;
	const int kVerLo = 0;
#pragma pack(1)
	struct tFlatHeader
	{
		unsigned int fVersion;
		float fRadius;
		int fTotalLODs;
		float fLODHysteresis;
	}flatHeader = { (kVerHi << 8) | (kVerLo & 0xFF), mGetBoundingRadius(), modelCount, gHysteresis };
#pragma pack()
	fwrite( &flatHeader, sizeof( tFlatHeader ), 1, myFile );
	fwrite( &gLODTransitions, sizeof(float), kMaxLOD, myFile );

	////////////////////////
	// Strings STRS chunk
	//
	if ( fStringPool.Count() )
	{
		tChunk stringChunk;
		stringChunk.mStart( myFile, MAKEFOURCC( 'S', 'T', 'R', 'S' ) );
		fwrite( fStringPool.Addr(0), fStringPool.Count(), 1, myFile );
		stringChunk.mFixUp();
	}

	////////////////////////
	// Named Points PNTS chunk
	//
	if ( fNumNamedPoints )
	{
		tChunk pointsChunk;
		pointsChunk.mStart( myFile, MAKEFOURCC( 'P', 'N', 'T', 'S' ) );
		fwrite( fNamedPoints, sizeof(tNamedPoint), fNumNamedPoints, myFile );
		pointsChunk.mFixUp();
	}

	////////////////////////
	// Material MATS chunk
	//
	debugMsg( "MATS chunk" );
	tChunk materialChunk;
	materialChunk.mStart( myFile, MAKEFOURCC( 'M', 'A', 'T', 'S' ) );
#pragma pack(1)
	struct tMaterial
	{
		enum
		{
			kTexMapped      = 0x0001,	// set if texture mapping
			kGouraud        = 0x0002,	// set if gouraud (smooth) shading
			kFlat			= 0x0004,	// set if flat shading
			kWire           = 0x0008,	// set if polygon should be wire frame
			kAdditive		= 0x0010,	// set if polygon should be blended
			kLuminance		= 0x0020	// set if we have a luminance texture map
		};
		unsigned int fFlags;
		COLORREF fColor;
		float fAmbient;
		float fDiffuse;
		float fEmittance;
	}material = { 0, RGB( 0xff, 0xff, 0xff ), 1.0f, 1.0f, 0.0f };
	struct tTextureInfo
	{
		int fTexID;
		int fTexNameOffset;
	}textureInfo[2] = { { 0, 0 }, { 0, 0 } };
#pragma pack()
	// Save one default material
	if ( fAddGenericMaterial )
	{
		material.fFlags = tMaterial::kFlat;
		fwrite( &material, sizeof(tMaterial), 1, myFile );
		fwrite( &textureInfo[0], sizeof(tTextureInfo), 1, myFile );
	}
	int count = 0;
	// Save actual material lists
	debugMsg( "Number of Max materials %d", myMtls.Count() );
	for ( i = 0; i < myMtls.Count(); i++ )
	{
		textureInfo[0].fTexNameOffset = 0;
		textureInfo[1].fTexNameOffset = 0;
		material.fFlags = ( fSmoothedMaterials & ( 1 << i ) ) ? tMaterial::kGouraud : tMaterial::kFlat;
		// Debug the Max material structures
		if( myMtls[i].m && myMtls[i].m->GetName().data() )
		{
			memset( &material, 0, sizeof(tMaterial) );
			bool gotLuminance = false;
			for ( int j = 0; j < myMtls[i].m->NumSubTexmaps(); j++ )
			{
				Texmap* map = myMtls[i].m->GetSubTexmap(j);
				if ( map && ( j == ID_DI || j == ID_SI ) )	// Get diffuse and self-illumination
				{
					// just do bitmap textures for now
					BitmapTex* bitmap = NULL;
					if ( map->ClassID() == Class_ID( BMTEX_CLASS_ID, 0 ) )
					{
						bitmap = (BitmapTex *) map;
						if ( map->GetName().data() && bitmap && bitmap->GetMapName() )
						{
							if ( j == ID_DI )
							{
								textureInfo[0].fTexNameOffset = mGetTextureNameIndex(count++);
							}
							else
							{
								textureInfo[1].fTexNameOffset = mGetTextureNameIndex(count++);
								gotLuminance = true;
							}
							material.fFlags = ( fSmoothedMaterials & ( 1 << i ) ) ? tMaterial::kGouraud : tMaterial::kFlat;
							material.fFlags |= tMaterial::kTexMapped;
						}
					}
				}
			}
			if ( gotLuminance )
			{
				material.fFlags |= tMaterial::kLuminance;
			}
			material.fAmbient = ( myMtls[i].m->GetAmbient().r + myMtls[i].m->GetAmbient().g + myMtls[i].m->GetAmbient().b ) / 3;
			material.fDiffuse = ( myMtls[i].m->GetDiffuse().r + myMtls[i].m->GetDiffuse().g + myMtls[i].m->GetDiffuse().b ) / 3;
			material.fEmittance = myMtls[i].m->GetSelfIllum();
			// Store the material definition
			material.fColor = RGB( 0xff, 0xff, 0xff );
			fwrite( &material, sizeof(tMaterial), 1, myFile );
			fwrite( &textureInfo[0], sizeof(tTextureInfo), ( (material.fFlags & tMaterial::kLuminance) ? 2 : 1 ), myFile );
		}
	}
	materialChunk.mFixUp();


	pMesh = fFirstMesh;
	while ( pMesh )
	{
		debugMsg( "MOD chunk" );
		///////////////////////
		// Model MOD chunk
		//

		tChunk modelChunk;
		modelChunk.mStart( myFile, MAKEFOURCC( 'M', 'O', 'D', '0' ) );

		fwrite( &pMesh->fNumVertices, sizeof(int), 1, myFile );
		fwrite( &pMesh->fNumFaces, sizeof(int), 1, myFile );

		debugMsg( "VTXS chunk" );
		// Output all of the vertices
		tChunk vertexChunk;
		vertexChunk.mStart( myFile, MAKEFOURCC( 'V', 'T', 'X', 'S' ) );
		fwrite( pMesh->fVertices, sizeof( tPoint3 ), pMesh->fNumVertices, myFile );
		vertexChunk.mFixUp();

		debugMsg( "POLY chunk" );
		// Save the Polygons
		for ( i = 0; i < pMesh->fNumFaces; i++ )
		{
			tChunk polyChunk;
			polyChunk.mStart( myFile, MAKEFOURCC( 'P', 'O', 'L', 'Y' ) );
			fwrite( &pMesh->fFaces[i], sizeof( tFaceDef ), 1, myFile );
			polyChunk.mFixUp();
		}

		// End Model MOD chunk
		//
		modelChunk.mFixUp();

		pMesh = pMesh->fNext;
	}

	// End Main MDLS chunk
	mainChunk.mFixUp();

	// Close the file
	fflush( myFile );
	fclose( myFile );

	return 1;
}

int tSFCMODExport::DoExport(const TCHAR *filename,ExpInterface *ei,Interface *gi, BOOL suppressPrompts, DWORD options) 
{
	// be back MessageBox( GetActiveWindow(), "tSFCMODExport::DoExport", GetString( IDS_SD_SFCMODEXPERROR ),  MB_OK );
	debugMsg( "tSFCMODExport::DoExport" );
//	DialogBox( hInstance, MAKEINTRESOURCE( IDD_GREETING ), GetActiveWindow(), MsgDlgProc );

	// Set a global prompt display switch
	showPrompts = suppressPrompts ? FALSE : TRUE;
	exportSelected = (options & SCENE_EXPORT_SELECTED) ? TRUE : FALSE;

	int status;

	status = mSaveSFCMOD( filename, ei, gi );

	if(status == 0)
		return 1;		// Dialog cancelled
	if(status < 0)
		return 0;		// Real, honest-to-goodness error
	return(status);
}

BOOL tSFCMODExport::SupportsOptions(int ext, DWORD options) 
{
	assert(ext == 0);	// We only support one extension
	return(options == SCENE_EXPORT_SELECTED) ? TRUE : FALSE;
}

